#include "BF_MSNR.h"

#include <stdio.h>
#include <iostream>

namespace BF
{
    CMSNR::CMSNR(int _numFreq, int _samplingRate, double _soundSpeed, std::string _decomp_alg, double _alpha) :
	BF(_numFreq, _samplingRate, _soundSpeed),
	decomp_alg(_decomp_alg), alpha(_alpha)
    {
    }

    CMSNR::~CMSNR()
    {
    }

    void
    CMSNR::updateSources(const std::vector<SourceLocation> &sourcePositions)
    {
/*== source from adaptBF.m

  % This algorithm requires VAD data (Adata.vad(NSrc), = 0(noise) or 1(signal))
  % This algorithm uses 2 intermediate variables Rss, Rnn, mean 
  % correlation matrixes of signal(incl. noise) and noise, respectively.
  % The freedom of scaling is fixed by |g| = 1. (g is BF coef. vector)
  % This algorithm requires to set initial G for getting size information.
  % To set zero matrix for G is OK.
      if isempty(Adata.Rss) % Initialization process
        Adata.Rss = zeros(NMic, NMic, NSrc, NFreq); % Signal correlation matrix
        Adata.Rnn = zeros(NMic, NMic, NSrc, NFreq); % Noise correlation matrix
        for nf = 1:NFreq
          for ns = 1:NSrc % for each source
            Adata.Rss(:, :, ns, nf) = eye(NMic);
            Adata.Rnn(:, :, ns, nf) = eye(NMic);
            G(ns, :, nf) = ones(NMic, 1)./sqrt(NMic); % Initial BF coef.
          end
        end
        return
      else % update process
        for ns = 1:NSrc % for each source
          vad = Adata.vad(ns);
          for nf = 1:NFreq
              x = Adata.x(:, nf);
              Rxx = x*x';
              alp = Adata.alp;
              Rss = Adata.Rss(:, :, ns, nf);
              Rnn = Adata.Rnn(:, :, ns, nf);
              if vad == 1 % Signal part
                  Rss = alp * Rss + (1-alp) * Rxx;
                  Adata.Rss(:, :, ns, nf) = Rss;
              else % Noise part
                  Rnn = alp * Rnn + (1-alp) * Rxx;
                  Adata.Rnn(:, :, ns, nf) = Rnn;
              end
              [V, D] = eig(Rss, Rnn);
              [msnr, mi] = max(diag(D));
              g = V(:, mi);
              g = g ./ sqrt(real(g'*g));
              G(ns, :, nf) = g'; % Store total BF coef.
          end
        end
        return
      end % For if initialization
      
*/
	const int numSrc = sourcePositions.size();

	/*== 直近フレームでの音源と、このフレームでの音源との対応をとる ==*/
	std::vector<int> lastSourceIndices; // i 番めの要素は、このフレームでの i 番目の音源（配列の i 番目の要素という意味）が、直近フレームでの何番目の音源（同）に相当するかを表す。
	for(int i = 0; i < sourcePositions.size(); i++){
	    const SourceLocation &sourcePosition = sourcePositions[i];

	    /* 直近フレームの音源から、同じ ID を持つものを探す */
	    int j;
	    for(j = 0; j < this->sourcePositions.size(); j++){
		if(sourcePosition.sourceID == this->sourcePositions[j].sourceID){
		    break;
		}
	    }
	    
	    if(j < this->sourcePositions.size()){
		/* 見つかった */
		lastSourceIndices.push_back(j);
	    }else{
		/* 見つからなかった */
		lastSourceIndices.push_back(-1);
	    }
	}

	/*== 伝達関数・分離行列を再計算する ==*/
	//std::vector<Matrix> newHs;
	std::vector<Matrix> newGs;
	//this->calculateTransformFunctions(micPositions, sourcePositions, newHs);

	const int numMic = this->numMicChannels;
	
	//MSNRからマイク数を与える必要がある。マイク数はinputでわかる

	std::vector<Matrix> newRss, newRnn, newG;
	std::vector<int> newRssCnt, newRnnCnt;
	for(int i = 0; i < this->numFreq; i++){

	    Matrix newG(numSrc, numMic);
	    for(int j = 0; j < numSrc; j++){
		const SourceLocation &srcPos = sourcePositions[j];

		/* 入力の j 番目の音源と同じIDの音源が、直近フレームに見当たらない？ */
		const int lastSourceIndex = lastSourceIndices[j];
		
		if(lastSourceIndex < 0){
#if 1
		    newRss.push_back(Matrix::I(numMic));
		    newRnn.push_back(Matrix::I(numMic));
#else
		    newRssCnt.push_back(0);
		    newRss.push_back(Matrix::O(numMic,numMic));
		    newRnnCnt.push_back(0);
		    newRnn.push_back(Matrix::O(numMic,numMic));
#endif
		    for(int k = 0; k < numMic; k++){
			newG.setElementAt(j, k, 1.0 / sqrt(numMic));
		    }
		}else{
#if 1
		    newRss.push_back(this->Rss[i * this->sourcePositions.size() + lastSourceIndex]);
		    newRnn.push_back(this->Rnn[i * this->sourcePositions.size() + lastSourceIndex]);
#else
		    newRss.push_back(this->Rss[i * this->sourcePositions.size() + lastSourceIndex]);
		    newRnn.push_back(this->Rnn[i * this->sourcePositions.size() + lastSourceIndex]);
		    newRssCnt.push_back(this->RssCnt[i * this->sourcePositions.size() + lastSourceIndex]);
		    newRnnCnt.push_back(this->RnnCnt[i * this->sourcePositions.size() + lastSourceIndex]);
#endif
		    newG.  setRow(j, this->separationMatrices[i].getRow(lastSourceIndex));
		}
	    }

	    newGs.push_back(newG);
	}

	/*== 結果の格納 ==*/
	this->sourcePositions = sourcePositions;
	newRss.swap(this->Rss);
	newRnn.swap(this->Rnn);
#if 1
#else
	newRssCnt.swap(this->RssCnt);
	newRnnCnt.swap(this->RnnCnt);
#endif
	newGs.swap(this->separationMatrices);
    }

    void
    CMSNR::adaptSeparationMatrices()
    {
	const int numSrc = this->sourcePositions.size();
	//const int numMic = this->transformFunctions[0].getRowCount();
	const int numMic = this->numMicChannels;
	
	std::vector<Matrix> newRss, newRnn, newGs;
	for(int i = 0; i < this->numFreq; i++){
	    
	    Matrix newG(numSrc, numMic);
	    for(int j = 0; j < numSrc; j++){
		bool isSignal = this->sourcePositions[j].isSignal;
		//fprintf(stdout, "isSignal[%d] = %d\n", j, isSignal);

		const Matrix &G = this->separationMatrices[i];
	    
		Matrix x     = this->inputSpectrum.getColumn(i);
		Matrix rxx   = x * x.transposed();
		double alpha = this->alpha;
		
		Matrix rss, rnn;
		Matrix rsst, rnnt;
		if(isSignal){
#if 1
		    rss = alpha * this->Rss[i * numSrc + j] + (1 - alpha) * rxx;
		    rnn = this->Rnn[i * numSrc + j];
#else
		    RssCnt[i * numSrc + j] = RssCnt[i * numSrc + j] + 1;
		    rss = this->Rss[i * numSrc + j] + rxx;
		    rnn = this->Rnn[i * numSrc + j];
		    rsst = this->Rss[i * numSrc + j] * (1.0 / RssCnt[i * numSrc + j]) + rxx * (1.0 / RssCnt[i * numSrc + j]);
		    if(RnnCnt[i * numSrc + j] == 0)
		      rnnt = Matrix::I(numMic);
		    else
		      rnnt = this->Rnn[i * numSrc + j] * (1.0 / RnnCnt[i * numSrc + j]);
#endif
		}else{
#if 1
		    rss = this->Rss[i * numSrc + j];
		    rnn = alpha * this->Rnn[i * numSrc + j] + (1 - alpha) * rxx;
#else
		    RnnCnt[i * numSrc + j] = RnnCnt[i * numSrc + j] + 1;
		    rss = this->Rss[i * numSrc + j];
		    rnn = this->Rnn[i * numSrc + j] + rxx;
		    if(RssCnt[i * numSrc + j] == 0)
		      rsst = Matrix::I(numMic);
		    else
		      rsst = this->Rss[i * numSrc + j] * ( 1.0 / RssCnt[i * numSrc + j]);
		    rnnt = this->Rnn[i * numSrc + j] * ( 1.0 / RnnCnt[i * numSrc + j]) + rxx * (1.0 / RnnCnt[i * numSrc + j]);
#endif
		}
		//rss.dump(stdout, "rss[%d,%d]", i, j);
		//rnn.dump(stdout, "rnn[%d,%d]", i, j);
		newRnn.push_back(rnn);
		newRss.push_back(rss);

		Matrix evec, eval, right_evec;
		/* Rnn が I でない場合、octave の結果とこの一般固有値の計算結果は一致しないようである。
		 * ただし、octave の結果は AV-BVD が 0 ではないので、octave 実装に誤差があると予想する。
		 */
#if 1
		if(decomp_alg == "GEVD")
		  rss.generalizedEigenValues(rnn, eval, evec);
		else if(decomp_alg == "GSVD")
		  rss.generalizedSingularValues(rnn, evec, eval, right_evec);
#else
		if(decomp_alg == "GEVD")
		  rsst.generalizedEigenValues(rnnt, eval, evec);
		else if(decomp_alg == "GSVD")
		  rsst.generalizedSingularValues(rnnt, evec, eval, right_evec);
#endif
		//evec.dump(stdout, "evec");
		//eval.dump(stdout, "eval");

		//(rss * evec).dump(stdout, "rss * evec");
		//(rnn * evec * eval.diagonalMatrix()).dump(stdout, "rnn * evec");
		//((rss * evec).elementwiseDivided(rnn * evec)).dump(stdout, "(rss * evec)./(rnn * evec)");

		int m = eval.getMaximumValueIndexInColumn(0);
		Matrix g = evec.getColumn(m);
		//fprintf(stdout, "m=%d\n", m);
		//g.dump(stdout, "g");
		g = 1 / g.norm().real() * g;
		//g.dump(stdout, "g");
		newG.setRow(j, g.transposed());
	    }
	    newGs.push_back(newG);
	}

	// std::cout << RssCnt[0] << " " << RnnCnt[0] << std::endl;

	//newRss[2].dump(stdout, "Rss[2]");
	//newRnn[2].dump(stdout, "Rnn[2]");
	//newGs[0].dump(stdout, "G[0]");

	/*== 結果の格納 ==*/
	newRss.swap(this->Rss);
	newRnn.swap(this->Rnn);
	newGs.swap(this->separationMatrices);
    }
}
